/*
 *  linux/fs/nfs/nfs4proc.c
 *
 *  nfs version 4 client side proceedure stub
 *
 *  by Andy Adamson <andros@citi.umich.edu>
 *  Copyright (C) 1999 The Regents of the University of Michigan
 *
 */


#define NFS_NEED_NFS4_XDR_TYPES

#include <linux/param.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/utsname.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/in.h>
#include <linux/pagemap.h>
#include <linux/sunrpc/clnt.h>
#include <linux/nfs.h>
#include <linux/nfs4_prot.h>
#include <linux/nfs_fs.h>

#include <asm/segment.h>

/* Global State */

clientid4 nfs4_clientid = 0; 			/* set at mount*/
verifier4 nfs_verf;				/*set at init */

int
nfs4_proc_compound(struct nfs_server *server, COMPOUND4args *cargs, COMPOUND4res *cres)
{
	int status;

	nfsv4_printk(compound,"NFSV4 call compound\n");
	nfsv4_printk(kendrick, "nfs4_proc_compound() : starting with tag \"%.*s\"\n",
		     (int) cargs->tag.utf8string_len, cargs->tag.utf8string_val);

	status = nfs4_set_bsize(server->client, NFSPROC4_COMPOUND, cargs, cres);
	if(status != 0)
		goto out;

	status = rpc_call(server->client, NFSPROC4_COMPOUND,cargs,cres,0);
	nfsv4_printk (kendrick, "nfs4_proc_compound() : returning rescnt %d status %d\n", cres->RESCNT, status);
	nfsv4_printk(compound,"NFSV4 reply compound. status: %d \n",status);

 out:
/* the NFSV4 protocol always returns a positive status. the rpc layer
 * returns a negative status. always present the nfsv4 client with a
 * positive status
 */
	if(status > 0)
		return status;
	return -status;
}

struct nfs4_rpc_ops	nfs_v4_clientops = {
	4,			/* protocol version */
	nfs4_proc_compound,
};
