/*
 * linux/fs/nfsd/nfsd4util.c
 *
 * debug print facilities
 *
 * by Andy Adamson <andros@citi.umich.edu>
 * Copyright (C) 1999 The Regents of the University of Michigan
 */

#include <linux/fs.h>

#include <linux/sunrpc/svc.h>
#include <linux/nfsd/nfsd4_prot.h>


#define NFSDDBG_FACILITY		NFSDDBG_PROC

/* nfsv4_printk toggles */
int nfsv4_debug_attributes      = 0;    /* fattr4, including xdr */
int nfsv4_debug_fh4             = 0;    /* file handle, including xdr */
int nfsv4_debug_utf8            = 1;    /* strings , including xdr*/
int nfsv4_debug_compound        = 1;    /* compound call */
int nfsv4_debug_xdr             = 1;    /* generic xdr: u64,bitmap4,specdata4 */
int nfsv4_debug_vfs             = 1;    /* generic vfs  (vfs.c)*/
int nfsv4_debug_export          = 1;    /* generic export (export.c) */
int nfsv4_debug_fhops           = 0;    /* generic fhops  (nfsfh.c)*/
int nfsv4_debug_svc             = 1;    /* generic svc  (nfssvc.c)*/
int nfsv4_debug_free            = 0;    /* kfree (nfsd4xdr.c)*/
int nfsv4_debug_n2s             = 0;    /* number to string cache(n2s_cache.c)*/
int nfsv4_debug_open            = 0;    /* OPEN call */
int nfsv4_debug_close           = 0;    /* CLOSE call */
int nfsv4_debug_rename          = 0;    /* RENAME call */
int nfsv4_debug_link		= 0;		/* LINK call */
int nfsv4_debug_bitmap		= 0;		/* set bitmap  */
int nfsv4_debug_secinfo		= 1;		/* SECINFO4 */
int nfsv4_debug_setclntid	= 1;		/* SETCLIENTID*/
int nfsv4_debug_pseudo		= 0;		/* pseudo.c  */
int nfsv4_debug_kendrick        = 1;
int nfsv4_debug_clid            = 1;
int nfsv4_debug_andros          = 1;
int nfsv4_debug_lease           = 0;
int nfsv4_debug_lock            = 1;
int nfsv4_debug_write           = 0;
int nfsv4_debug_read            = 0;
int nfsv4_debug_callback        = 0;
int nfsv4_debug_deleg           = 0;
int nfsv4_debug_dispatcher	= 0;

void
print_rpcsec_gss_info(rpcsec_gss_info *ptr)
{
	int i;
	char *p;

	nfsv4_printk(secinfo,"NFSV4: print_rpcsec_gss_info\n");
	/*secinfo4*/
		
	nfsv4_printk(secinfo,
		     "NFSV4: sec_info4 len: %d\n sec_info4 val:",ptr->oid.sec_oid4_len);
	p = ptr->oid.sec_oid4_val;
	for(i=0;i<ptr->oid.sec_oid4_len;i++)
		nfsv4_printk(secinfo,"%c",*p++);
	nfsv4_printk(secinfo,"NFSV4: qop: %d\n",ptr->qop);
	nfsv4_printk(secinfo,"NFSV4: service: %d\n",ptr->service);
}

void
print_verifier4(verifier4 v1)
{
	nfsv4_printk(setclntid,"%x\n",*(unsigned int *)&v1[0]);
	nfsv4_printk(setclntid,"%x\n",*(unsigned int *)&v1[4]);
}

void
print_fattr4(fattr4 *fptr)
{
	int i=0;
	int bit;
	bitmap4 *bm4 = &fptr->attrmask;
	char *cptr = fptr->FATTRSVAL;
	int flen = fptr->FATTRSLEN;
	char *end = cptr + flen;

	if(!bm4)
		return;
	/* for each bit set in the bit mask */
	nfsv4_printk(attributes,"NFSV4:print_fattr4 bitmap len %d \n",bm4->BLEN);
	nfsv4_printk(attributes,"NFSV4:print_fattr4 attrlist len %d \n",fptr->FATTRSLEN);
	nfsv4_printk(attributes,"NFSV4:print_fattr4 end - cptr %d \n",end - cptr);

	do{
		bit = get_bitmap4_bit(bm4,i);
		i++;
		nfsv4_printk(attributes,"NFSV4:print_fattr4 get_bitmap4_bit returns %d \n",bit);
		if(!bit || bit == -1)
			continue;


		nfsv4_printk(attributes,
			     "NFSV4:print_fattr4 bit %d is set\n",i);
		switch(i-1){
		case FATTR4_TYPE:
		{
			nfsv4_printk(attributes,
				     "        TYPE: %d\n", *(fattr4_type *)cptr);
			cptr += sizeof(fattr4_type);
		}
		break;
		case FATTR4_SIZE:
		{
	
			nfsv4_printk(attributes,"        SIZE: %d\n", *(u32 *)cptr);
			cptr += sizeof(fattr4_size);
		}
		break;
		case FATTR4_FSID:
		{
				/*fattr4_fsid *ptr = (fattr4_fsid *)cptr; */
			u32 *iptr = (u32 *)cptr;
			nfsv4_printk(attributes,"        FSID Major: 0x%x:0x%x  FSID Minor: 0x%x:0x%x\n",*iptr++,*iptr++,*iptr++,*iptr);
			cptr += sizeof(fattr4_fsid);
		}
		break;
		case FATTR4_FILEID:
		{
			nfsv4_printk(attributes,"        FILEID: %ld\n",*(long int *)cptr);
			cptr += sizeof(fattr4_fileid);
		}
		break;
		case FATTR4_MODE:
		{
			nfsv4_printk(attributes,"        MODE: %d\n",*(fattr4_mode *)cptr);
			cptr += sizeof(fattr4_mode);
		}
		break;
		case FATTR4_NUMLINKS:
		{
			nfsv4_printk(attributes,"        NUMLINKS: %d\n", *(fattr4_numlinks *)cptr);
			cptr += sizeof(fattr4_numlinks);
		}
		break;
		case FATTR4_RAWDEV:
		{
			u32 *iptr = (u32 *)cptr;
			nfsv4_printk(attributes,"        RAWDEV specdata1: 0x%x  specdata2: 0x%x\n",*iptr++,*iptr);
			cptr += sizeof(fattr4_rawdev);
		}
		break;
		case FATTR4_SPACE_USED:
		{
			nfsv4_printk(attributes,"        SPACE_USED: %ld\n", *(long int *)cptr);
			cptr += sizeof(fattr4_space_used);
		}
		break;
		case FATTR4_TIME_ACCESS:
		{
			fattr4_time_access *ptr = (fattr4_time_access *)cptr;
			nfsv4_printk(attributes,"        TIME_ACCESS: %ld\n",(long int)ptr->seconds);
			cptr += sizeof(fattr4_time_access);
		}
		break;
		case FATTR4_TIME_CREATE:
		{
			fattr4_time_create *ptr = (fattr4_time_create *)cptr;
			nfsv4_printk(attributes,"        TIME_CREATE: %ld\n",(long int)ptr->seconds);
			cptr += sizeof(fattr4_time_create);
		}
		break;
		case FATTR4_TIME_MODIFY:
		{
			fattr4_time_modify *ptr = (fattr4_time_modify *)cptr;
			nfsv4_printk(attributes,"        TIME_MODIFY: %ld\n",(long int)ptr->seconds);
			cptr += sizeof(fattr4_time_modify);
		}
		break;
		}
	}while(bit != -1);
}

void
print_nfs_fh4(nfs_fh4 *fh4)
{
	char *kptr= fh4->nfs_fh4_val;
	int i;

	nfsv4_printk(fh4,"NFSV4 print_nfs_fh4 ");
	nfsv4_printk(fh4,"nfs_fh4_val: 0x%x\n",(unsigned int)kptr);
	if(fh4->nfs_fh4_len <= 32){
		for(i=0;i<fh4->nfs_fh4_len;i++){
			nfsv4_printk(fh4,"%x",*kptr);
			kptr++;
		}
		nfsv4_printk(fh4,"\n");
	}else
		nfsv4_printk(fh4," nfs_fh4_len too big: %d\n",fh4->nfs_fh4_len);
}

void
print_knfs_fh(struct knfs_fh *kfh)
{
	char *kptr= (char *)kfh;
	int i,size = sizeof(struct knfs_fh);

	nfsv4_printk(fh4,"NFSV4 print_knfs_fh size: %d\n",size);
	for(i=0;i<size;i++){
		nfsv4_printk(fh4,"%x",*kptr);
		kptr++;
	}
	nfsv4_printk(fh4,"\n");
	nfsv4_printk(fh4,"fh_base 0x%x, ino %d dirino %d dev %d xino %d xdev %d\n",
		     (unsigned int)&kfh->fh_base,	
		     kfh->fh_base.fb_ino,
		     kfh->fh_base.fb_dirino,
		     kfh->fh_base.fb_dev,
		     kfh->fh_base.fb_xino,
		     kfh->fh_base.fb_xdev);
}


void
print_utf8string(utf8string *str)
{
	char *ptr = str->utf8string_val;
	int i;

	for(i=0;i<str->utf8string_len;i++){
		nfsv4_printk(utf8,"%c",*ptr);
		ptr++;
	}
	nfsv4_printk(utf8,"\n");
}

void
dump_utf8string(utf8string *str)
{
	char *ptr = str->utf8string_val;
	int i;

	for(i=0;i<str->utf8string_len;i++){
		printk("%c",*ptr);
		ptr++;
	}
	nfsv4_printk(utf8,"\n");
}

void
hex_dump (char *p, int no_bytes)
{
	int i;

	for (i = 0; i < no_bytes; i++) {
		printk ("%x", (unsigned char) p[i]);

		if ((i % 56) == 55)
			printk ("\n\n");
		else if ((i % 8) == 7)
			printk ("\n");
		else if ((i % 4) == 3)
			printk ("   ");
		else
			printk (" ");
	}
	printk ("\n");
}
