/*
 *  linux/include/linux/nfs4_prot.h
 *
 *  nfs version 4 protocol definitions
 *
 *  by Andy Adamson <andros@citi.umich.edu>
 *  Copyright (C) 1999 The Regents of the University of Michigan
 *
 */

/*
 *  This file was initially generated using /usr/bin/rpcgen on the
 *  ietf nfs_prot.x. Because kernel and user xdr definitions are different,
 *  a stuff has been changed, and a lot has been added.
 */

#ifndef _NFS4_PROT_H_RPCGEN
#define	_NFS4_PROT_H_RPCGEN

#ifdef __linux__
#include <linux/sunrpc/clnt.h>
#include <linux/sunrpc/svc.h>
#include <linux/nfs_fs_sb.h>
#endif /* __linux__ */

#ifdef __OpenBSD__
#include <nfs/nfs_fs_sb.h>
#define GFP_KERNEL 0

void *kmalloc(unsigned long s, int t);
void kfree(void *p);

typedef uint8_t u8;
typedef uint16_t u16;
typedef uint32_t u32;
typedef uint32_t __u32;
typedef uint64_t u64;
typedef uint64_t __u64;
typedef dev_t kdev_t;

#define printk(args...) printf(args)
#define memcpy(d, s, l) bcopy((s), (d), (l))

#endif /* OpenBSD */

#define rpcauth_lookupcred(x,y)	(NULL)

#ifdef NFSV4_DEBUG
#define ifon(fac)       (nfsv4_debug_##fac == 1)
#define nfsv4_printk(fac,args...)  do{ifon(fac)?printk(args):0;}while(0)
#else /*NFSV4_DEBUG*/
#define nfsv4_printk(fac,args...)  do{}while(0);
#endif /*NFSV4_DEBUG*/

/* debug printk toggles, see nfsd4proc.c */
extern int nfsv4_debug_attributes;      /* fattr4, including xdr*/
extern int nfsv4_debug_fh4;             /* file handle, including xdr*/
extern int nfsv4_debug_utf8;            /* strings, including xdr */
extern int nfsv4_debug_compound;        /* compound call */
extern int nfsv4_debug_xdr;             /* generic xdr */
extern int nfsv4_debug_vfs;             /* generic vfs (vfs.c)*/
extern int nfsv4_debug_export;          /* generic export (export.c)*/
extern int nfsv4_debug_fhops;           /* generic fhops (nfsfh.c)*/
extern int nfsv4_debug_svc;             /* generic svc (nfssvc.c)*/
extern int nfsv4_debug_dir;             /* readdir (dir.c)*/
extern int nfsv4_debug_v4tov3;          /* translation functions (nfs4proc.c)*/
extern int nfsv4_debug_inode;           /* inode (inode.c)*/
extern int nfsv4_debug_read;            /* read (read.c)*/
extern int nfsv4_debug_readdir;         /* READDIR */
extern int nfsv4_debug_write;           /* write (write.c)*/
extern int nfsv4_debug_free;            /* kfree (nfs4_util.c)*/
extern int nfsv4_debug_open;            /* OPEN call (inode.c) */
extern int nfsv4_debug_create;          /* nfs_create */
extern int nfsv4_debug_close;           /* CLOSE call (inode.c) */
extern int nfsv4_debug_rename;          /* RENAME call */
extern int nfsv4_debug_link;            /* LINK call */
extern int nfsv4_debug_n2s;             /* num2str driver */
extern int nfsv4_debug_secinfo;         /* SECINFO */
extern int nfsv4_debug_setclntid;       /* SETCLIENTID */
extern int nfsv4_debug_fileops;       /* unimplemented calls in file.c */
extern int nfsv4_debug_bsize;         /* bsize calculations */
extern int nfsv4_debug_setattr;					/* SETATTR */
extern int nfsv4_debug_lock;
extern int nfsv4_debug_verifier;
extern int nfsv4_debug_andros;
extern int nfsv4_debug_kendrick;
extern int nfsv4_debug_callback;
extern int nfsv4_debug_deleg;

#define NFS4_MAXNAMLEN          NAME_MAX
#define	NFS4_FHSIZE 128
#define	NFS4_VERIFIER_SIZE 8

/* from RFC 2203 */
#define RPCSEC_GSS  6


#define AUTH_NONE 0
#define AUTH_UNIX 1

typedef int bool_t;

/* the maximum number of entry4's allowed per readdir
 * this is needed as an overrun check, and is a guess.
 * and will be removed as soon as i clear up READDIR4.
 */
#define NFS4_MAX_DIR            PAGE_SIZE/32  /*4096/32 = 256*/

/* added NF4BAD, NF4NON */
enum nfs_ftype4 {
	NF4NON = 0,
	NF4REG = 1,
	NF4DIR = 2,
	NF4BLK = 3,
	NF4CHR = 4,
	NF4LNK = 5,
	NF4SOCK = 6,
	NF4FIFO = 7,
	NF4ATTRDIR = 8,
	NF4NAMEDATTR = 9,
	NF4BAD = 10
};
typedef enum nfs_ftype4 nfs_ftype4;

enum nfsstat4 {
	NFS4_OK = 0,
	NFS4ERR_PERM = 1,
	NFS4ERR_NOENT = 2,
	NFS4ERR_IO = 5,
	NFS4ERR_NXIO = 6,
	NFS4ERR_ACCES = 13,
	NFS4ERR_EXIST = 17,
	NFS4ERR_XDEV = 18,
	NFS4ERR_NODEV = 19,
	NFS4ERR_NOTDIR = 20,
	NFS4ERR_ISDIR = 21,
	NFS4ERR_INVAL = 22,
	NFS4ERR_FBIG = 27,
	NFS4ERR_NOSPC = 28,
	NFS4ERR_ROFS = 30,
	NFS4ERR_MLINK = 31,
	NFS4ERR_NAMETOOLONG = 63,
	NFS4ERR_NOTEMPTY = 66,
	NFS4ERR_DQUOT = 69,
	NFS4ERR_STALE = 70,
	NFS4ERR_BADHANDLE = 10001,
	NFS4ERR_BAD_COOKIE = 10003,
	NFS4ERR_NOTSUPP = 10004,
	NFS4ERR_TOOSMALL = 10005,
	NFS4ERR_SERVERFAULT = 10006,
	NFS4ERR_BADTYPE = 10007,
	NFS4ERR_DELAY = 10008,
	NFS4ERR_SAME = 10009,
	NFS4ERR_DENIED = 10010,
	NFS4ERR_EXPIRED = 10011,
	NFS4ERR_LOCKED = 10012,
	NFS4ERR_GRACE = 10013,
	NFS4ERR_FHEXPIRED = 10014,
	NFS4ERR_SHARE_DENIED = 10015,
	NFS4ERR_WRONGSEC = 10016,
	NFS4ERR_CLID_INUSE = 10017,
	NFS4ERR_RESOURCE = 10018,
	NFS4ERR_MOVED = 10019,
	NFS4ERR_NOFILEHANDLE = 10020,
	NFS4ERR_MINOR_VERS_MISMATCH = 10021,
	NFS4ERR_STALE_CLIENTID = 10022,
	NFS4ERR_STALE_STATEID = 10023,
	NFS4ERR_OLD_STATEID = 10024,
	NFS4ERR_BAD_STATEID = 10025,
	NFS4ERR_BAD_SEQID = 10026,
	NFS4ERR_NOT_SAME = 10027,
	NFS4ERR_LOCK_RANGE = 10028,
	NFS4ERR_SYMLINK = 10029,
	NFS4ERR_READDIR_NOSPC = 10030,
	NFS4ERR_LEASE_MOVED = 10031,
};
typedef enum nfsstat4 nfsstat4;

#ifdef __OpenBSD__
struct rpc_rqst {
	struct iovec *rq_svec;
	struct iovec *rq_rvec;
	uint32_t rq_slen;
	uint32_t rq_rlen;
};

#define xdr_adjust_iovec(iov, p) \
    ((iov)->iov_len = ((uint8_t *) (p) - (uint8_t *) ((iov)->iov_base)))
#endif

typedef struct {
	u_int bitmap4_len;
	uint32_t *bitmap4_val;
} bitmap4;
#define BLEN bitmap4_len
#define BVAL bitmap4_val

typedef uint64_t offset4;

typedef uint32_t count4;

typedef uint64_t length4;

typedef uint64_t clientid4;

typedef uint64_t stateid4;

typedef uint32_t seqid4;

typedef struct {
	u_int utf8string_len;
	char *utf8string_val;
} utf8string;

typedef utf8string component4;

typedef struct {
	u_int pathname4_len;
	component4 *pathname4_val;
} pathname4;

typedef uint64_t nfs_lockid4;

typedef uint64_t nfs_cookie4;

typedef utf8string linktext4;

typedef struct {
	u_int sec_oid4_len;
	char *sec_oid4_val;
} sec_oid4;

typedef uint32_t qop4;

typedef uint32_t mode4;

typedef uint64_t changeid4;

typedef char verifier4[NFS4_VERIFIER_SIZE];

struct nfstime4 {
	int64_t seconds;
	uint32_t nseconds;
};
typedef struct nfstime4 nfstime4;

enum time_how4 {
	SET_TO_SERVER_TIME4 = 0,
	SET_TO_CLIENT_TIME4 = 1,
};
typedef enum time_how4 time_how4;

struct settime4 {
	time_how4 set_it;
	union {
		nfstime4 time;
	} settime4_u;
};
typedef struct settime4 settime4;

typedef struct {
	u_int nfs_fh4_len;
	char *nfs_fh4_val;
} nfs_fh4;

struct fsid4 {
	uint64_t major;
	uint64_t minor;
};
typedef struct fsid4 fsid4;

struct fs_location4 {
	struct {
		u_int server_len;
		utf8string *server_val;
	} server;
	pathname4 rootpath;
};
typedef struct fs_location4 fs_location4;

struct fs_locations4 {
	pathname4 fs_root;
	struct {
		u_int locations_len;
		fs_location4 *locations_val;
	} locations;
};
typedef struct fs_locations4 fs_locations4;
#define ACL4_SUPPORT_ALLOW_ACL 0x00000001
#define ACL4_SUPPORT_DENY_ACL 0x00000002
#define ACL4_SUPPORT_AUDIT_ACL 0x00000004
#define ACL4_SUPPORT_ALARM_ACL 0x00000008

typedef uint32_t acetype4;
#define ACE4_ACCESS_ALLOWED_ACE_TYPE 0x00000000
#define ACE4_ACCESS_DENIED_ACE_TYPE 0x00000001
#define ACE4_SYSTEM_AUDIT_ACE_TYPE 0x00000002
#define ACE4_SYSTEM_ALARM_ACE_TYPE 0x00000003

typedef uint32_t aceflag4;
#define ACE4_FILE_INHERIT_ACE 0x00000001
#define ACE4_DIRECTORY_INHERIT_ACE 0x00000002
#define ACE4_NO_PROPAGATE_INHERIT_ACE 0x00000004
#define ACE4_INHERIT_ONLY_ACE 0x00000008
#define ACE4_SUCCESSFUL_ACCESS_ACE_FLAG 0x00000010
#define ACE4_FAILED_ACCESS_ACE_FLAG 0x00000020
#define ACE4_IDENTIFIER_GROUP 0x00000040

typedef uint32_t acemask4;
#define ACE4_READ_DATA 0x00000001
#define ACE4_LIST_DIRECTORY 0x00000001
#define ACE4_WRITE_DATA 0x00000002
#define ACE4_ADD_FILE 0x00000002
#define ACE4_APPEND_DATA 0x00000004
#define ACE4_ADD_SUBDIRECTORY 0x00000004
#define ACE4_READ_NAMED_ATTRS 0x00000008
#define ACE4_WRITE_NAMED_ATTRS 0x00000010
#define ACE4_EXECUTE 0x00000020
#define ACE4_DELETE_CHILD 0x00000040
#define ACE4_READ_ATTRIBUTES 0x00000080
#define ACE4_WRITE_ATTRIBUTES 0x00000100
#define ACE4_DELETE 0x00010000
#define ACE4_READ_ACL 0x00020000
#define ACE4_WRITE_ACL 0x00040000
#define ACE4_WRITE_OWNER 0x00080000
#define ACE4_SYNCHRONIZE 0x00100000
#define ACE4_GENERIC_READ 0x00120081
#define ACE4_GENERIC_WRITE 0x00160106
#define ACE4_GENERIC_EXECUTE 0x001200A0

struct nfsace4 {
	acetype4 type;
	aceflag4 flag;
	acemask4 access_mask;
	utf8string who;
};
typedef struct nfsace4 nfsace4;

struct specdata4 {
	uint32_t specdata1;
	uint32_t specdata2;
};
typedef struct specdata4 specdata4;
#define FH4_PERSISTENT 0x00000000
#define FH4_NOEXPIRE_WITH_OPEN 0x00000001
#define FH4_VOLATILE_ANY 0x00000002
#define FH4_VOL_MIGRATION 0x00000004
#define FH4_VOL_RENAME 0x00000008

typedef bitmap4 fattr4_supported_attrs;

typedef nfs_ftype4 fattr4_type;

typedef uint32_t fattr4_fh_expire_type;

typedef changeid4 fattr4_change;

typedef uint64_t fattr4_size;

typedef bool_t fattr4_link_support;

typedef bool_t fattr4_symlink_support;

typedef bool_t fattr4_named_attr;

typedef fsid4 fattr4_fsid;

typedef bool_t fattr4_unique_handles;

typedef uint32_t fattr4_lease_time;

typedef nfsstat4 fattr4_rdattr_error;

typedef struct {
	u_int fattr4_acl_len;
	nfsace4 *fattr4_acl_val;
} fattr4_acl;
#define FLEN    fattr4_acl_len
#define FVAL    fattr4_acl_val

typedef uint32_t fattr4_aclsupport;

typedef bool_t fattr4_archive;

typedef bool_t fattr4_cansettime;

typedef bool_t fattr4_case_insensitive;

typedef bool_t fattr4_case_preserving;

typedef bool_t fattr4_chown_restricted;

typedef uint64_t fattr4_fileid;

typedef uint64_t fattr4_files_avail;

typedef nfs_fh4 fattr4_filehandle;

typedef uint64_t fattr4_files_free;

typedef uint64_t fattr4_files_total;

typedef fs_locations4 fattr4_fs_locations;

typedef bool_t fattr4_hidden;

typedef bool_t fattr4_homogeneous;

typedef uint64_t fattr4_maxfilesize;

typedef uint32_t fattr4_maxlink;

typedef uint32_t fattr4_maxname;

typedef uint64_t fattr4_maxread;

typedef uint64_t fattr4_maxwrite;

typedef utf8string fattr4_mimetype;

typedef mode4 fattr4_mode;

typedef bool_t fattr4_no_trunc;

typedef uint32_t fattr4_numlinks;

typedef utf8string fattr4_owner;

typedef utf8string fattr4_owner_group;

typedef uint64_t fattr4_quota_hard;

typedef uint64_t fattr4_quota_soft;

typedef uint64_t fattr4_quota_used;

typedef specdata4 fattr4_rawdev;

typedef uint64_t fattr4_space_avail;

typedef uint64_t fattr4_space_free;

typedef uint64_t fattr4_space_total;

typedef uint64_t fattr4_space_used;

typedef bool_t fattr4_system;

typedef nfstime4 fattr4_time_access;

typedef settime4 fattr4_time_access_set;

typedef nfstime4 fattr4_time_backup;

typedef nfstime4 fattr4_time_create;

typedef nfstime4 fattr4_time_delta;

typedef nfstime4 fattr4_time_metadata;

typedef nfstime4 fattr4_time_modify;

typedef settime4 fattr4_time_modify_set;
#define FATTR4_SUPPORTED_ATTRS 0
#define FATTR4_TYPE 1
#define FATTR4_FH_EXPIRE_TYPE 2
#define FATTR4_CHANGE 3
#define FATTR4_SIZE 4
#define FATTR4_LINK_SUPPORT 5
#define FATTR4_SYMLINK_SUPPORT 6
#define FATTR4_NAMED_ATTR 7
#define FATTR4_FSID 8
#define FATTR4_UNIQUE_HANDLES 9
#define FATTR4_LEASE_TIME 10
#define FATTR4_RDATTR_ERROR 11
#define FATTR4_ACL 12
#define FATTR4_ACLSUPPORT 13
#define FATTR4_ARCHIVE 14
#define FATTR4_CANSETTIME 15
#define FATTR4_CASE_INSENSITIVE 16
#define FATTR4_CASE_PRESERVING 17
#define FATTR4_CHOWN_RESTRICTED 18
#define FATTR4_FILEHANDLE 19
#define FATTR4_FILEID 20
#define FATTR4_FILES_AVAIL 21
#define FATTR4_FILES_FREE 22
#define FATTR4_FILES_TOTAL 23
#define FATTR4_FS_LOCATIONS 24
#define FATTR4_HIDDEN 25
#define FATTR4_HOMOGENEOUS 26
#define FATTR4_MAXFILESIZE 27
#define FATTR4_MAXLINK 28
#define FATTR4_MAXNAME 29
#define FATTR4_MAXREAD 30
#define FATTR4_MAXWRITE 31
#define FATTR4_MIMETYPE 32
#define FATTR4_MODE 33
#define FATTR4_NO_TRUNC 34
#define FATTR4_NUMLINKS 35
#define FATTR4_OWNER 36
#define FATTR4_OWNER_GROUP 37
#define FATTR4_QUOTA_HARD 38
#define FATTR4_QUOTA_SOFT 39
#define FATTR4_QUOTA_USED 40
#define FATTR4_RAWDEV 41
#define FATTR4_SPACE_AVAIL 42
#define FATTR4_SPACE_FREE 43
#define FATTR4_SPACE_TOTAL 44
#define FATTR4_SPACE_USED 45
#define FATTR4_SYSTEM 46
#define FATTR4_TIME_ACCESS 47
#define FATTR4_TIME_ACCESS_SET 48
#define FATTR4_TIME_BACKUP 49
#define FATTR4_TIME_CREATE 50
#define FATTR4_TIME_DELTA 51
#define FATTR4_TIME_METADATA 52
#define FATTR4_TIME_MODIFY 53
#define FATTR4_TIME_MODIFY_SET 54

/* not in spec. */
#define FATTR4_ATTR_MAX   55
#define XDR_ATTR4_BLEN  (FATTR4_ATTR_MAX + 32)/32

typedef struct {
	u_int attrlist4_len;
	char *attrlist4_val;
} attrlist4;

struct fattr4 {
	bitmap4 attrmask;
	attrlist4 attr_vals;
};
typedef struct fattr4 fattr4;
#define FATTRSLEN attr_vals.attrlist4_len
#define FATTRSVAL attr_vals.attrlist4_val

struct change_info4 {
	bool_t atomic;
	changeid4 before;
	changeid4 after;
};
typedef struct change_info4 change_info4;

struct clientaddr4 {
	char *r_netid;
	char *r_addr;
};
typedef struct clientaddr4 clientaddr4;

struct cb_client4 {
	u_int cb_program;
	clientaddr4 cb_location;
};
typedef struct cb_client4 cb_client4;

struct nfs_client_id4 {
	verifier4 verifier;
	struct {
		u_int id_len;
		char *id_val;
	} id;
};
typedef struct nfs_client_id4 nfs_client_id4;

struct nfs_lockowner4 {
	clientid4 clientid;
	struct {
		u_int owner_len;
		char *owner_val;
	} owner;
};
typedef struct nfs_lockowner4 nfs_lockowner4;

enum nfs_lock_type4 {
	READ_LT = 1,
	WRITE_LT = 2,
	READW_LT = 3,
	WRITEW_LT = 4,
};
typedef enum nfs_lock_type4 nfs_lock_type4;
#define ACCESS4_READ 0x00000001
#define ACCESS4_LOOKUP 0x00000002
#define ACCESS4_MODIFY 0x00000004
#define ACCESS4_EXTEND 0x00000008
#define ACCESS4_DELETE 0x00000010
#define ACCESS4_EXECUTE 0x00000020

struct ACCESS4args {
	uint32_t access;
};
typedef struct ACCESS4args ACCESS4args;

struct ACCESS4resok {
	uint32_t supported;
	uint32_t access;
};
typedef struct ACCESS4resok ACCESS4resok;
#define SUPPORTED ACCESS4res_u.resok4.supported
#define ACCESS ACCESS4res_u.resok4.access

struct ACCESS4res {
	nfsstat4 status;
	union {
		ACCESS4resok resok4;
	} ACCESS4res_u;
};
typedef struct ACCESS4res ACCESS4res;

struct CLOSE4args {
	seqid4 seqid;
	stateid4 stateid;
};
typedef struct CLOSE4args CLOSE4args;

struct CLOSE4res {
	nfsstat4 status;
	union {
		stateid4 stateid;
	} CLOSE4res_u;
};
typedef struct CLOSE4res CLOSE4res;

struct COMMIT4args {
	offset4 offset;
	count4 count;
};
typedef struct COMMIT4args COMMIT4args;

struct COMMIT4resok {
	verifier4 writeverf;
};
typedef struct COMMIT4resok COMMIT4resok;

struct COMMIT4res {
	nfsstat4 status;
	union {
		COMMIT4resok resok4;
	} COMMIT4res_u;
};
typedef struct COMMIT4res COMMIT4res;

struct createtype4 {
	nfs_ftype4 type;
	union {
		linktext4 linkdata;
		specdata4 devdata;
	} createtype4_u;
};
typedef struct createtype4 createtype4;

struct CREATE4args {
	component4 objname;
	createtype4 objtype;
};
typedef struct CREATE4args CREATE4args;

struct CREATE4resok {
	change_info4 cinfo;
};
typedef struct CREATE4resok CREATE4resok;

struct CREATE4res {
	nfsstat4 status;
	union {
		CREATE4resok resok4;
	} CREATE4res_u;
};
typedef struct CREATE4res CREATE4res;
#define CREATEinfo CREATE4res_u.resok4.cinfo

struct DELEGPURGE4args {
	clientid4 clientid;
};
typedef struct DELEGPURGE4args DELEGPURGE4args;

struct DELEGPURGE4res {
	nfsstat4 status;
};
typedef struct DELEGPURGE4res DELEGPURGE4res;

struct DELEGRETURN4args {
	stateid4 stateid;
};
typedef struct DELEGRETURN4args DELEGRETURN4args;

struct DELEGRETURN4res {
	nfsstat4 status;
};
typedef struct DELEGRETURN4res DELEGRETURN4res;

struct GETATTR4args {
	bitmap4 attr_request;
};
typedef struct GETATTR4args GETATTR4args;

struct GETATTR4resok {
	fattr4 obj_attributes;
};
typedef struct GETATTR4resok GETATTR4resok;
#define GFATTR4 GETATTR4res_u.resok4.obj_attributes

struct GETATTR4res {
	nfsstat4 status;
	union {
		GETATTR4resok resok4;
	} GETATTR4res_u;
};
typedef struct GETATTR4res GETATTR4res;

struct GETFH4resok {
	nfs_fh4 object;
};
typedef struct GETFH4resok GETFH4resok;
#define GETFH4FH GETFH4res_u.resok4.object

struct GETFH4res {
	nfsstat4 status;
	union {
		GETFH4resok resok4;
	} GETFH4res_u;
};
typedef struct GETFH4res GETFH4res;

struct LINK4args {
	component4 newname;
};
typedef struct LINK4args LINK4args;

struct LINK4resok {
	change_info4 cinfo;
};
typedef struct LINK4resok LINK4resok;
#define LINK_RES_INFO LINK4res_u.resok4.cinfo

struct LINK4res {
	nfsstat4 status;
	union {
		LINK4resok resok4;
	} LINK4res_u;
};
typedef struct LINK4res LINK4res;

struct LOCK4args {
	nfs_lock_type4 locktype;
	seqid4 seqid;
	bool_t reclaim;
	stateid4 stateid;
	offset4 offset;
	length4 length;
};
typedef struct LOCK4args LOCK4args;

struct LOCK4denied {
	nfs_lockowner4 owner;
	offset4 offset;
	length4 length;
};
typedef struct LOCK4denied LOCK4denied;

struct LOCK4res {
	nfsstat4 status;
	union {
		stateid4 stateid;
		LOCK4denied denied;
	} LOCK4res_u;
};
typedef struct LOCK4res LOCK4res;

struct LOCKT4args {
	nfs_lock_type4 locktype;
	nfs_lockowner4 owner;
	offset4 offset;
	length4 length;
};
typedef struct LOCKT4args LOCKT4args;

struct LOCKT4res {
	nfsstat4 status;
	union {
		LOCK4denied denied;
	} LOCKT4res_u;
};
typedef struct LOCKT4res LOCKT4res;

struct LOCKU4args {
	nfs_lock_type4 type;
	seqid4 seqid;
	stateid4 stateid;
	offset4 offset;
	length4 length;
};
typedef struct LOCKU4args LOCKU4args;

struct LOCKU4res {
	nfsstat4 status;
	union {
		stateid4 stateid;
	} LOCKU4res_u;
};
typedef struct LOCKU4res LOCKU4res;

struct LOOKUP4args {
	pathname4 path;
};
typedef struct LOOKUP4args LOOKUP4args;
#define PLEN path.pathname4_len
#define PVAL path.pathname4_val

struct LOOKUP4res {
	nfsstat4 status;
};
typedef struct LOOKUP4res LOOKUP4res;

struct LOOKUPP4res {
	nfsstat4 status;
};
typedef struct LOOKUPP4res LOOKUPP4res;

struct NVERIFY4args {
	fattr4 obj_attributes;
};
typedef struct NVERIFY4args NVERIFY4args;

struct NVERIFY4res {
	nfsstat4 status;
};
typedef struct NVERIFY4res NVERIFY4res;

enum createmode4 {
	UNCHECKED4 = 0,
	GUARDED4 = 1,
	EXCLUSIVE4 = 2,
};
typedef enum createmode4 createmode4;

struct createhow4 {
	createmode4 mode;
	union {
		fattr4 createattrs;
		verifier4 createverf;
	} createhow4_u;
};
typedef struct createhow4 createhow4;
#define CREATE_VERF createhow4_u.createverf
#define CREATE_ATTR createhow4_u.createattrs

enum opentype4 {
	OPEN4_NOCREATE = 0,
	OPEN4_CREATE = 1,
};
typedef enum opentype4 opentype4;

struct openflag4 {
	opentype4 opentype;
	union {
		createhow4 how;
	} openflag4_u;
};
typedef struct openflag4 openflag4;
#define HOW_MODE openflag4_u.how.mode
#define HOW openflag4_u.how

enum limit_by4 {
	NFS_LIMIT_SIZE = 1,
	NFS_LIMIT_BLOCKS = 2,
};
typedef enum limit_by4 limit_by4;

struct nfs_modified_limit4 {
	uint32_t num_blocks;
	uint32_t bytes_per_block;
};
typedef struct nfs_modified_limit4 nfs_modified_limit4;

struct nfs_space_limit4 {
	limit_by4 limitby;
	union {
		uint64_t filesize;
		nfs_modified_limit4 mod_blocks;
	} nfs_space_limit4_u;
};
typedef struct nfs_space_limit4 nfs_space_limit4;
#define OPEN4_SHARE_ACCESS_READ 0x00000001
#define OPEN4_SHARE_ACCESS_WRITE 0x00000002
#define OPEN4_SHARE_ACCESS_BOTH 0x00000003
#define OPEN4_SHARE_DENY_NONE 0x00000000
#define OPEN4_SHARE_DENY_READ 0x00000001
#define OPEN4_SHARE_DENY_WRITE 0x00000002
#define OPEN4_SHARE_DENY_BOTH 0x00000003

enum open_delegation_type4 {
	OPEN_DELEGATE_NONE = 0,
	OPEN_DELEGATE_READ = 1,
	OPEN_DELEGATE_WRITE = 2,
};
typedef enum open_delegation_type4 open_delegation_type4;

enum open_claim_type4 {
	CLAIM_NULL = 0,
	CLAIM_PREVIOUS = 1,
	CLAIM_DELEGATE_CUR = 2,
	CLAIM_DELEGATE_PREV = 3,
};
typedef enum open_claim_type4 open_claim_type4;

struct open_claim_delegate_cur4 {
	pathname4 file;
	stateid4 delegate_stateid;
};
typedef struct open_claim_delegate_cur4 open_claim_delegate_cur4;

struct open_claim4 {
	open_claim_type4 claim;
	union {
		pathname4 file;
		uint32_t delegate_type;
		open_claim_delegate_cur4 delegate_cur_info;
		pathname4 file_delegate_prev;
	} open_claim4_u;
};
typedef struct open_claim4 open_claim4;
#define CLAIM_PATH_LEN open_claim4_u.file.pathname4_len
#define CLAIM_PATH_VAL open_claim4_u.file.pathname4_val

struct OPEN4args {
	open_claim4 claim;
	openflag4 openhow;
	nfs_lockowner4 owner;
	seqid4 seqid;
	uint32_t share_access;
	uint32_t share_deny;
};
typedef struct OPEN4args OPEN4args;
#define OWNER_LEN owner.owner.owner_len
#define OWNER_VAL owner.owner.owner_val

struct open_read_delegation4 {
	stateid4 stateid;
	bool_t recall;
	nfsace4 permissions;
};
typedef struct open_read_delegation4 open_read_delegation4;

struct open_write_delegation4 {
	stateid4 stateid;
	bool_t recall;
	nfs_space_limit4 space_limit;
	nfsace4 permissions;
};
typedef struct open_write_delegation4 open_write_delegation4;

struct open_delegation4 {
	open_delegation_type4 delegation_type;
	union {
		open_read_delegation4 read;
		open_write_delegation4 write;
	} open_delegation4_u;
};
typedef struct open_delegation4 open_delegation4;
#define OPEN4_RESULT_MLOCK 0x00000001
#define OPEN4_RESULT_CONFIRM 0x00000002

struct OPEN4resok {
	stateid4 stateid;
	change_info4 cinfo;
	uint32_t rflags;
	verifier4 open_confirm;
	open_delegation4 delegation;
};
typedef struct OPEN4resok OPEN4resok;
#define DEL_READ delegation.open_delegation4_u.read

struct OPEN4res {
	nfsstat4 status;
	union {
		OPEN4resok result;
	} OPEN4res_u;
};
typedef struct OPEN4res OPEN4res;
#define OPEN_RES OPEN4res_u.result

struct OPENATTR4res {
	nfsstat4 status;
};
typedef struct OPENATTR4res OPENATTR4res;

struct OPEN_CONFIRM4args {
	seqid4 seqid;
	verifier4 open_confirm;
};
typedef struct OPEN_CONFIRM4args OPEN_CONFIRM4args;

struct OPEN_CONFIRM4resok {
	stateid4 stateid;
};
typedef struct OPEN_CONFIRM4resok OPEN_CONFIRM4resok;

struct OPEN_CONFIRM4res {
	nfsstat4 status;
	union {
		OPEN_CONFIRM4resok resok4;
	} OPEN_CONFIRM4res_u;
};
typedef struct OPEN_CONFIRM4res OPEN_CONFIRM4res;
#define OPEN_CON_RES OPEN_CONFIRM4res_u.resok4

struct OPEN_DOWNGRADE4args {
	stateid4 stateid;
	seqid4 seqid;
	uint32_t share_access;
	uint32_t share_deny;
};
typedef struct OPEN_DOWNGRADE4args OPEN_DOWNGRADE4args;

struct OPEN_DOWNGRADE4resok {
	stateid4 stateid;
};
typedef struct OPEN_DOWNGRADE4resok OPEN_DOWNGRADE4resok;

struct OPEN_DOWNGRADE4res {
	nfsstat4 status;
	union {
		OPEN_DOWNGRADE4resok resok4;
	} OPEN_DOWNGRADE4res_u;
};
typedef struct OPEN_DOWNGRADE4res OPEN_DOWNGRADE4res;

struct PUTFH4args {
	nfs_fh4 object;
};
typedef struct PUTFH4args PUTFH4args;

struct PUTFH4res {
	nfsstat4 status;
};
typedef struct PUTFH4res PUTFH4res;

struct PUTPUBFH4res {
	nfsstat4 status;
};
typedef struct PUTPUBFH4res PUTPUBFH4res;

struct PUTROOTFH4res {
	nfsstat4 status;
};
typedef struct PUTROOTFH4res PUTROOTFH4res;

struct READ4args {
	stateid4 stateid;
	offset4 offset;
	count4 count;
/* iovec and nr do not go on the wire */
	struct iovec iov[3];
	int nr;
};
typedef struct READ4args READ4args;

struct READ4resok {
	bool_t eof;
	struct {
		u_int data_len;
		char *data_val;
	} data;
};
typedef struct READ4resok READ4resok;

struct READ4res {
	nfsstat4 status;
	union {
		READ4resok resok4;
	} READ4res_u;
};
typedef struct READ4res READ4res;
#define READLEN READ4res_u.resok4.data.data_len
#define READVAL READ4res_u.resok4.data.data_val
#define READEOF READ4res_u.resok4.eof
#define READDATA READ4res_u.resok4.data

struct READDIR4args {
	nfs_cookie4 cookie;
	verifier4 cookieverf;
	count4 dircount;
	count4 maxcount;
	bitmap4 attr_request;
/* iovec and nr do not go on the wire */
	struct iovec iov[3];
	int nr;
};
typedef struct READDIR4args READDIR4args;

struct entry4 {
	nfs_cookie4 cookie;
	component4 name;
	fattr4 attrs;
	struct entry4 *nextentry;
};
typedef struct entry4 entry4;

struct dirlist4 {
	entry4 *entries;
	bool_t eof;
};
typedef struct dirlist4 dirlist4;

struct READDIR4resok {
	verifier4 cookieverf;
	dirlist4 reply;
};
typedef struct READDIR4resok READDIR4resok;
#define DIR4 READDIR4res_u.resok4.reply
#define REOF READDIR4res_u.resok4.reply.eof
#define ENTRIES READDIR4res_u.resok4.reply.entries

struct READDIR4res {
	nfsstat4 status;
	union {
		READDIR4resok resok4;
	} READDIR4res_u;
};
typedef struct READDIR4res READDIR4res;

struct READLINK4resok {
	linktext4 link;
};
typedef struct READLINK4resok READLINK4resok;

struct READLINK4res {
	nfsstat4 status;
	union {
		READLINK4resok resok4;
	} READLINK4res_u;
};
typedef struct READLINK4res READLINK4res;

struct REMOVE4args {
	component4 target;
};
typedef struct REMOVE4args REMOVE4args;

struct REMOVE4resok {
	change_info4 cinfo;
};
typedef struct REMOVE4resok REMOVE4resok;

struct REMOVE4res {
	nfsstat4 status;
	union {
		REMOVE4resok resok4;
	} REMOVE4res_u;
};
typedef struct REMOVE4res REMOVE4res;

struct RENAME4args {
	component4 oldname;
	component4 newname;
};
typedef struct RENAME4args RENAME4args;
#define OLDNAME_VAL oldname.utf8string_val
#define NEWNAME_VAL newname.utf8string_val
#define OLDNAME_LEN oldname.utf8string_len
#define NEWNAME_LEN newname.utf8string_len

struct RENAME4resok {
	change_info4 source_cinfo;
	change_info4 target_cinfo;
};
typedef struct RENAME4resok RENAME4resok;

struct RENAME4res {
	nfsstat4 status;
	union {
		RENAME4resok resok4;
	} RENAME4res_u;
};
typedef struct RENAME4res RENAME4res;
#define RENAME_RES_SRC RENAME4res_u.resok4.source_cinfo
#define RENAME_RES_TAR RENAME4res_u.resok4.target_cinfo

struct RENEW4args {
	stateid4 stateid;
};
typedef struct RENEW4args RENEW4args;

struct RENEW4res {
	nfsstat4 status;
};
typedef struct RENEW4res RENEW4res;

struct RESTOREFH4res {
	nfsstat4 status;
};
typedef struct RESTOREFH4res RESTOREFH4res;

struct SAVEFH4res {
	nfsstat4 status;
};
typedef struct SAVEFH4res SAVEFH4res;

struct SECINFO4args {
	component4 name;
};
typedef struct SECINFO4args SECINFO4args;

#define RPC_GSS_SVC_T
enum rpc_gss_svc_t {
	RPC_GSS_SVC_NONE = 1,
	RPC_GSS_SVC_INTEGRITY = 2,
	RPC_GSS_SVC_PRIVACY = 3,
};
typedef enum rpc_gss_svc_t rpc_gss_svc_t;

struct rpcsec_gss_info {
	sec_oid4 oid;
	qop4 qop;
	rpc_gss_svc_t service;
};
typedef struct rpcsec_gss_info rpcsec_gss_info;

struct secinfo4 {
	uint32_t flavor;
	struct {
		u_int flavor_info_len;
		char *flavor_info_val;
	} flavor_info;
};
typedef struct secinfo4 secinfo4;

typedef struct {
	u_int SECINFO4resok_len;
	secinfo4 *SECINFO4resok_val;
} SECINFO4resok;

struct SECINFO4res {
	nfsstat4 status;
	union {
		SECINFO4resok resok4;
	} SECINFO4res_u;
};
typedef struct SECINFO4res SECINFO4res;
#define SECRES SECINFO4res_u.resok4

struct SETATTR4args {
	stateid4 stateid;
	fattr4 obj_attributes;
};
typedef struct SETATTR4args SETATTR4args;

struct SETATTR4res {
	nfsstat4 status;
	bitmap4 attrsset;
};
typedef struct SETATTR4res SETATTR4res;

struct SETCLIENTID4args {
	nfs_client_id4 client;
	cb_client4 callback;
};
typedef struct SETCLIENTID4args SETCLIENTID4args;

struct SETCLIENTID4resok {
	clientid4 clientid;
	verifier4 setclientid_confirm;
};
typedef struct SETCLIENTID4resok SETCLIENTID4resok;

struct SETCLIENTID4res {
	nfsstat4 status;
	union {
		SETCLIENTID4resok resok4;
		clientaddr4 client_using;
	} SETCLIENTID4res_u;
};
typedef struct SETCLIENTID4res SETCLIENTID4res;
#define CLIDOK SETCLIENTID4res_u.resok4

struct SETCLIENTID_CONFIRM4args {
	verifier4 setclientid_confirm;
};
typedef struct SETCLIENTID_CONFIRM4args SETCLIENTID_CONFIRM4args;

struct SETCLIENTID_CONFIRM4res {
	nfsstat4 status;
};
typedef struct SETCLIENTID_CONFIRM4res SETCLIENTID_CONFIRM4res;

struct VERIFY4args {
	fattr4 obj_attributes;
};
typedef struct VERIFY4args VERIFY4args;

struct VERIFY4res {
	nfsstat4 status;
};
typedef struct VERIFY4res VERIFY4res;

enum stable_how4 {
	UNSTABLE4 = 0,
	DATA_SYNC4 = 1,
	FILE_SYNC4 = 2,
};
typedef enum stable_how4 stable_how4;

struct WRITE4args {
	stateid4 stateid;
	offset4 offset;
	stable_how4 stable;
	struct {
		u_int data_len;
		char *data_val;
	} data;
};
typedef struct WRITE4args WRITE4args;
#define WRITELEN data.data_len
#define WRITEVAL data.data_val

struct WRITE4resok {
	count4 count;
	stable_how4 committed;
	verifier4 writeverf;
};
typedef struct WRITE4resok WRITE4resok;

struct WRITE4res {
	nfsstat4 status;
	union {
		WRITE4resok resok4;
	} WRITE4res_u;
};
typedef struct WRITE4res WRITE4res;
#define WRITERES WRITE4res_u.resok4

enum nfs_opnum4 {
	OP_ACCESS = 3,
	OP_CLOSE = 4,
	OP_COMMIT = 5,
	OP_CREATE = 6,
	OP_DELEGPURGE = 7,
	OP_DELEGRETURN = 8,
	OP_GETATTR = 9,
	OP_GETFH = 10,
	OP_LINK = 11,
	OP_LOCK = 12,
	OP_LOCKT = 13,
	OP_LOCKU = 14,
	OP_LOOKUP = 15,
	OP_LOOKUPP = 16,
	OP_NVERIFY = 17,
	OP_OPEN = 18,
	OP_OPENATTR = 19,
	OP_OPEN_CONFIRM = 20,
	OP_OPEN_DOWNGRADE = 21,
	OP_PUTFH = 22,
	OP_PUTPUBFH = 23,
	OP_PUTROOTFH = 24,
	OP_READ = 25,
	OP_READDIR = 26,
	OP_READLINK = 27,
	OP_REMOVE = 28,
	OP_RENAME = 29,
	OP_RENEW = 30,
	OP_RESTOREFH = 31,
	OP_SAVEFH = 32,
	OP_SECINFO = 33,
	OP_SETATTR = 34,
	OP_SETCLIENTID = 35,
	OP_SETCLIENTID_CONFIRM = 36,
	OP_VERIFY = 37,
	OP_WRITE = 38,
};
typedef enum nfs_opnum4 nfs_opnum4;

struct nfs_argop4 {
	nfs_opnum4 argop;
	union {
		ACCESS4args opaccess;
		CLOSE4args opclose;
		COMMIT4args opcommit;
		CREATE4args opcreate;
		DELEGPURGE4args opdelegpurge;
		DELEGRETURN4args opdelegreturn;
		GETATTR4args opgetattr;
		LINK4args oplink;
		LOCK4args oplock;
		LOCK4args oplockt;
		LOCK4args oplocku;
		LOOKUP4args oplookup;
		NVERIFY4args opnverify;
		OPEN4args opopen;
		OPEN_CONFIRM4args opopen_confirm;
		OPEN_DOWNGRADE4args opopen_downgrade;
		PUTFH4args opputfh;
		READ4args opread;
		READDIR4args opreaddir;
		REMOVE4args opremove;
		RENAME4args oprename;
		RENEW4args oprenew;
		SECINFO4args opsecinfo;
		SETATTR4args opsetattr;
		SETCLIENTID4args opsetclientid;
		SETCLIENTID_CONFIRM4args opsetclientid_confirm;
		VERIFY4args opverify;
		WRITE4args opwrite;
	} nfs_argop4_u;
};
typedef struct nfs_argop4 nfs_argop4;

struct nfs_resop4 {
	nfs_opnum4 resop;
	union {
		ACCESS4res opaccess;
		CLOSE4res opclose;
		COMMIT4res opcommit;
		CREATE4res opcreate;
		DELEGPURGE4res opdelegpurge;
		DELEGRETURN4res opdelegreturn;
		GETATTR4res opgetattr;
		GETFH4res opgetfh;
		LINK4res oplink;
		LOCK4res oplock;
		LOCKT4res oplockt;
		LOCKU4res oplocku;
		LOOKUP4res oplookup;
		LOOKUPP4res oplookupp;
		NVERIFY4res opnverify;
		OPEN4res opopen;
		OPENATTR4res opopenattr;
		OPEN_CONFIRM4res opopen_confirm;
		OPEN_DOWNGRADE4res opopen_downgrade;
		PUTFH4res opputfh;
		PUTPUBFH4res opputpubfh;
		PUTROOTFH4res opputrootfh;
		READ4res opread;
		READDIR4res opreaddir;
		READLINK4res opreadlink;
		REMOVE4res opremove;
		RENAME4res oprename;
		RENEW4res oprenew;
		RESTOREFH4res oprestorefh;
		SAVEFH4res opsavefh;
		SECINFO4res opsecinfo;
		SETATTR4res opsetattr;
		SETCLIENTID4res opsetclientid;
		SETCLIENTID_CONFIRM4res opsetclientid_confirm;
		VERIFY4res opverify;
		WRITE4res opwrite;
	} nfs_resop4_u;
};
typedef struct nfs_resop4 nfs_resop4;

struct COMPOUND4args {
	utf8string tag;
	uint32_t minorversion;
	struct {
		u_int argarray_len;
		nfs_argop4 *argarray_val;
	} argarray;
};
typedef struct COMPOUND4args COMPOUND4args;
#define OPVAL argarray.argarray_val
#define OPCNT argarray.argarray_len
#define OPU nfs_argop4_u

struct COMPOUND4res {
	nfsstat4 status;
	utf8string tag;
	struct {
		u_int resarray_len;
		nfs_resop4 *resarray_val;
	} resarray;
};
typedef struct COMPOUND4res COMPOUND4res;
#define RESCNT resarray.resarray_len
#define RESVAL resarray.resarray_val
#define RESU nfs_resop4_u

struct CB_GETATTR4args {
	nfs_fh4 fh;
	bitmap4 attr_request;
};
typedef struct CB_GETATTR4args CB_GETATTR4args;

struct CB_GETATTR4resok {
	fattr4 obj_attributes;
};
typedef struct CB_GETATTR4resok CB_GETATTR4resok;

struct CB_GETATTR4res {
	nfsstat4 status;
	union {
		CB_GETATTR4resok resok4;
	} CB_GETATTR4res_u;
};
typedef struct CB_GETATTR4res CB_GETATTR4res;

struct CB_RECALL4args {
	stateid4 stateid;
	bool_t truncate;
	nfs_fh4 fh;
};
typedef struct CB_RECALL4args CB_RECALL4args;

struct CB_RECALL4res {
	nfsstat4 status;
};
typedef struct CB_RECALL4res CB_RECALL4res;

enum nfs_cb_opnum4 {
	OP_CB_GETATTR = 3,
	OP_CB_RECALL = 4,
};
typedef enum nfs_cb_opnum4 nfs_cb_opnum4;

struct nfs_cb_argop4 {
	u_int argop;
	union {
		CB_GETATTR4args opcbgetattr;
		CB_RECALL4args opcbrecall;
	} nfs_cb_argop4_u;
};
typedef struct nfs_cb_argop4 nfs_cb_argop4;

struct nfs_cb_resop4 {
	u_int resop;
	union {
		CB_GETATTR4res opcbgetattr;
		CB_RECALL4res opcbrecall;
	} nfs_cb_resop4_u;
};
typedef struct nfs_cb_resop4 nfs_cb_resop4;

struct CB_COMPOUND4args {
	utf8string tag;
	uint32_t minorversion;
	struct {
		u_int argarray_len;
		nfs_cb_argop4 *argarray_val;
	} argarray;
};
typedef struct CB_COMPOUND4args CB_COMPOUND4args;

struct CB_COMPOUND4res {
	nfsstat4 status;
	utf8string tag;
	struct {
		u_int resarray_len;
		nfs_cb_resop4 *resarray_val;
	} resarray;
};
typedef struct CB_COMPOUND4res CB_COMPOUND4res;

#define NFS4_PROGRAM 100003
#define NFS_V4 4

#define NFSPROC4_NULL 0
#define NFSPROC4_COMPOUND 1
#define NFS4_CALLBACK 40000000
#define NFS_CB 1

#define CB_NULL 0
#define CB_COMPOUND 1

#ifdef __linux__
#include <linux/nfs4lock.h>
#endif /* __linux__ */

/* function prototypes */
#ifdef __linux__
/*
int nfs4_proc_compound(struct nfs_server *server, struct rpc_cred *cred, struct COMPOUND4args *cargs, struct COMPOUND4res *cres,int flags, rpc_action callback,  void *clntdata);
*/
int nfs4_proc_compound(struct nfs_server *server,struct COMPOUND4args *cargs, struct COMPOUND4res *cres);
#else
int nfs4_proc_compound(struct vnode *vp, struct proc *procp, struct ucred *cred, COMPOUND4args *cargs, COMPOUND4res *cres);
#endif

int nfs4_get_sizes(u32 *sendsizep, u32 *recvsizep, struct COMPOUND4args *args, struct COMPOUND4res *res);

int nfs4_set_bsize(struct rpc_clnt *clnt, u32 proc, struct COMPOUND4args *cargs, struct COMPOUND4res *cres);

void print_utf8string(utf8string *str);
int get_fattr4_res_buf_size(bitmap4 *bm4);
void stuffutf8(utf8string *utf8, char *str);
int setup_bitmap4(bitmap4 *bm,int len);
int set_bitmap4_bit(int b,bitmap4 *bm);
int nfs4_setupv3_getattr(GETATTR4args *cargs,GETATTR4res *cres);
void print_nfs_fh4(nfs_fh4 *fh4);
void print_nfs_fh(struct nfs_fh *fh);
void nfs4_setup_fhputfh(PUTFH4args *cargs, void *fhp, int fhsize);
void nfs4_setup_putfh(PUTFH4args *cargs,struct dentry *dentry);
void print_fattr4(fattr4 *ptr);
void setup_readdir_res(READDIR4args *args,READDIR4res *res);
void nfs4_stuff_nfsfattr(fattr4 *fptr,struct nfs_fattr *fattr);
int nfs4_stuff_v3dentry(nfs_fh4 *v4_fh,struct dentry *dentry);
#ifdef __OpenBSD__
int nfs4_stuff_v3fh(nfs_fh4 *v4_fh, nfsfh_t *v3_fh);
#else
int nfs4_stuff_v3fh(nfs_fh4 *v4_fh,struct nfs_fh *v3_fh);
void nfs4_stuff_nfsfsinfo(fattr4 *fptr,struct nfs_fsinfo *fsinfo);
int nfs4_setup_open (OPEN4args *cargs, struct dentry *dentry, int flag,
		     int mode, struct nfs_server *server);
int do_secinfo(struct inode *inode,struct dentry *dentry);
int nfs4_prepare_mount(struct super_block *sb,struct dentry *rdentry,struct nfs_fattr *fattr, struct nfs_fsinfo *fsinfo,struct nfs_fh *fh);
#endif
void qstr_to_lookup4(struct qstr *dname, LOOKUP4args *args);
void nfs4_kfree_compound_res(COMPOUND4res *res);
void nfs4_kfree_compound_args(COMPOUND4args *args);
int parse_mnt_path(char *mnt_path, LOOKUP4args *args);
void n2scache_init(void);
void nfs4_setup_rename(RENAME4args *args, struct qstr *old_name, struct qstr *new_name);
void nfs4_setup_symlink(CREATE4args *args, struct qstr *sym_name, struct qstr *dentry_name);
void nfs4_setup_mkdir(CREATE4args *args, struct qstr *dir_name);
int get_pathname4_bufsize(pathname4 *p4);
void nfs4_setup_link(LINK4args *args, struct qstr *new_name);
void nfs4_setup_remove(REMOVE4args *args, struct qstr *target_name);
void print_rpcsec_gss_info(rpcsec_gss_info *ptr);
int utf8cmp(utf8string *s1,utf8string *s2);
void utf8copy(utf8string *dest, utf8string *src);
void setup_compound(int opcnt,COMPOUND4args *cargs,COMPOUND4res *cres, char *tag,nfs_argop4 *aop4,nfs_resop4 *rop4);
int nfs4_xdr_enc_void(struct rpc_rqst *req, u32 *p, void *dummy);
int nfs4_xdr_compoundargs(struct rpc_rqst *req, u32 *p, struct COMPOUND4args *cargs);
int nfs4_xdr_dec_void(struct rpc_rqst *req, u32 *p, void *dummy);
int nfs4_xdr_compoundres(struct rpc_rqst *req, u32 *p, struct COMPOUND4res *cres);

struct super_block *get_empty_super(void);
struct vfsmount *add_vfsmnt(struct super_block *sb, const char *dev_name, const char *dir_name);
int get_lookup_arg_size(char *path);
int get_getattr4_size(int len);
int get_pathname4_size(char *path);
int get_compound_arg_size(char *str);
int get_putrootfh_arg_size(void);
int get_getfh_arg_size(void);

kdev_t get_unnamed_dev(void);
void verf_init(void);
void set_verifier(verifier4 target,verifier4 src);
int cmp_verifier(verifier4 v1, verifier4 v2);
void print_verifier4(verifier4 v1);
void nfs4_kfree_bitmap4(bitmap4 *bm4);
void nfs4_kfree_fh4(nfs_fh4 *fh4);
void nfs4_kfree_utf8(utf8string *utf8);
int do_setclientid(struct nfs_server *server);
void init_lock_state(void);
struct nfs4_lock_owner *nfs4_get_lockowner(pid_t pid,clientid4 clid);
int save_stateid(struct file *filp,stateid4 stateid);
void release_stateid(struct file *filep);
void shutdown_lock_state(void);
void n2s_cache_shutdown(void);
int nfs4_setup_close(CLOSE4args *args, clientid4 clid, stateid4 stateid);
int nfs4err2errno(int nfs4err);
void print_lockowner(nfs_lockowner4 *own);

extern int nfs4_setup_getattr (GETATTR4args *args, GETATTR4res *res, bitmap4 *bm);
extern void nfs4_setup_readdir (READDIR4args *args, READDIR4res *res, nfs_cookie4 cookie, bitmap4 *req, struct iovec *iov);
#ifdef __linux__
extern int nfs4_setup_setattr (SETATTR4args *args, SETATTR4res *res, struct iattr *attr);
extern int iattr_to_fattr4 (struct iattr *attr, fattr4 *fattr);
u32 check_open_confirm(OPEN4res *res,struct nfs_server *server,struct dentry *dentry);
#endif /* linux */
extern void dump_fattr4 (fattr4 *fptr);
extern void dump_OPEN4args (OPEN4args *argp);
extern void hex_dump (char *p, int no_bytes);

extern int xdr_decode_nullcb_args(struct svc_rqst *rqstp, u32 *p, void *data);
extern int xdr_encode_nullcb_res(struct svc_rqst *rqstp, u32 *p, void *data);
extern int xdr_free_nullcb_res(struct svc_rqst *rqstp, u32 *p, void *data);
extern int xdr_decode_compoundcb_args (struct svc_rqst *rqstp, u32 *p, CB_COMPOUND4args *argp);
extern int xdr_encode_compoundcb_res (struct svc_rqst *rqstp, u32 *p, CB_COMPOUND4res *resp);
extern void xdr_free_compoundcb_args (CB_COMPOUND4args *argp);
extern int xdr_free_compoundcb_res(struct svc_rqst *rqstp, u32 *p, CB_COMPOUND4res *data);
extern int nfs4_deleg_init(void);
extern void nfs4_setup_cb_client4(cb_client4 *clp);
extern void nfs4_accept_delegation(struct dentry *dentry, open_delegation4 *deleg);
extern int nfs4_delegreturn(struct inode *inode);
	
extern bitmap4 standard_bm;           // bitmaps for use with nfs4_setup_getattr()
	extern bitmap4 no_owner_bm;
	extern bitmap4 only_owner_bm;
	extern bitmap4 readdir_bm;
	extern bitmap4 statfs_bm;

#endif /* !_NFSD4_PROT_H_RPCGEN */
