/* linux/include/linux/nfs4lock.h 
 *
 * based on:
 * linux/include/linux/lockd/lockd.h
 * General-purpose lockd include file.
 * Copyright (C) 1996 Olaf Kirch <okir@monad.swb.de>
 *
 *
 */
                                                   
#ifndef _NFS4_LOCK_H
#define _NFS4_LOCK_H

#define NFS4_FILE(file)	((struct nfs4_file *)(file)->private_data)
#define NFS4_FILE_MAGIC 0xFEED

/*
 * lock owners
 * 
 * a lock owner is a clientid4/pid pair.
 * hash on clientid4. 
 */
struct nfs4_lock_owner {
	struct nfs4_lock_owner * lo_next;   /* linked list (hash table) */
	clientid4 lo_clientid;
	pid_t	lo_pid;
	unsigned int    lo_seq_num;  /* lock idempodent sequence number */
};

/* 
* filep private data. 
* note: the stateid might not be able to go into the file private data. 
* i will make that choice after delegation is implemented. 
* this struct occupies dentry->d_fsdata for a moment: 
* nfs_create() calls OPEN which in nfsv4 really opens the file
* and returns a stateid in d_fsdata which ends up in the filep in
* the subsequent nfs_open call. - see ../namei.c filp_open().
*/
struct nfs4_file {
	unsigned long f_magic;
	stateid4 f_stateid;			 /* returned by server, opaque to client */

	/* delegation info */
	open_delegation_type4 f_deleg_type;
	unsigned int f_fileflags;
	struct nfs4_file *f_next;
	struct nfs4_file **f_pprev;
};                                                             

static inline int
nfs4_check_file_stateid(struct file *file)
{
	if(!file || !file->private_data || (NFS4_FILE(file)->f_magic != NFS4_FILE_MAGIC)){
		printk(KERN_ERR "NFS: corrupt dentry file lock state!\n");
		return 0;
	}
	return 1;
}
#endif /* !_NFSD4_LOCK_H */
