/* linux/include/linux/nfsd/nfsdlock.h 
 *
 * based on:
 * linux/include/linux/lockd/lockd.h
 * General-purpose lockd include file.
 * Copyright (C) 1996 Olaf Kirch <okir@monad.swb.de>
 *
 *
 */
                                                   
#ifndef _NFSD4_LOCK_H
#define _NFSD4_LOCK_H
/*
 * This struct describes a file held open by lockd on behalf of
 * an NFS client.
 */
struct nfs4_file {
	struct nfs4_file * f_next;   /* linked list */
	struct knfs_fh    f_handle; /* NFS file handle */
	struct file   f_file;   /* VFS file pointer */
	struct nfs4_lock *  f_lock4; /* lock */
	unsigned int    f_count;  /* reference count incremented on open, decremented on close */
	struct semaphore  f_sema;   /* avoid concurrent access */
	unsigned int hashindex; /* to save recalculation */
	u_short	f_seq; /* unique accros nfs4file[hash] */
	u_short f_lock4_seq; /* sets lock-> l_generation. unique across this file's f_lock4s */

	struct semaphore	f_delegsema;     /* to serialize access to the file's delegation state */
	open_delegation_type4	f_delegtype;
	struct nfs4_delegation	*f_deleglist;
	int			f_wait_count;
	struct wait_queue	*f_wait_queue;
};                                               

struct nfs4_file_element {
	struct nfs4_file *f_hash_head;
	u_short f_uniq_seq; /* sets file->f_seq. unique per file in hash bucket */
	unsigned short  f_pad; /* word boundry */
};

/* nfs4_lock_owner lo_state values. zero indicates new lockowner */
#define OPEN_CONFIRM_REQ_NEGOTIATING	1
#define OPEN_CONFIRM_REQ_CONFIRMED		2

/*
 * lock owners
 * 
 */
struct nfs4_lock_owner {
	struct nfs4_lock_owner	* lo_next;		/* linked list (hash table) */
	struct nfs_lockowner4	*lo_owner;
	unsigned int		lo_seq_num;		/* last seqid4 received */
	int			lo_state;		/* open_confirmed state */
	struct nfs4_lock	*confirm_lock;		/* first OPEN: used for OPEN_CONFIRM */
	unsigned int		lo_confirm_seq_num;	/* first OPEN seqid4 received */
	struct svc_buf		lo_seq_response;	/* xdr response associated with lo_seq_num */
	struct semaphore	lo_seqid_sema;		/* must be held while processing a seq. number mutating op */
};

/*
 * nfs4_lock binds the lockowner to a file - expressed in stateid.
 *
 * holds share and record lock info including
 * per file ordered list of pending blocking locks.  
 * (well, not yet...)
 * NFSv4 does not support callbacks
 * on the release of a blocking lock - the client must poll.
 */
#define NLM_NEVER   (~(unsigned long) 0)

struct nfs4_lock {
  struct nfs4_lock *  l_next;   /* linked list off file->f_lock4*/
  struct nfs4_lock_owner * l_owner;              /* lock_owner */
  struct nfs4_file * l_file;  /* file in question */
	stateid4 l_stateid;			/* 32 bits of nfsd_verf, 
					                  * 16 bits of (nfs4_file) f_seq, 
					                  * 16 bits of (nfs4_lock) l_generation */
  u32 l_share_access; /* access mode */
  u32 l_share_deny;   /* deny mode */ 
	unsigned short	l_generation; /* uniqifier - used as part of the stateid */
	unsigned short  l_pad; /* word boundry */
};                                                             


extern __inline__ struct inode *
nfs4_file_inode(struct nfs4_file *file)
{
  return file->f_file.f_dentry->d_inode;
}

/*
 *  More file state needed for delegation
 */

struct nfs4_delegation {
	stateid4 d_stateid;
	struct svc_client *d_clnt;
	struct nfs4_file *d_file;
	struct knfs_fh d_fh;
	struct nfs4_delegation *d_next;
	struct nfs4_delegation *d_prev;
	struct nfs4_delegation *d_hashnext;
	struct nfs4_delegation *d_hashprev;
	int d_recalled;
};

#endif /* !_NFSD4_LOCK_H */
