/*
 * linux/include/linux/auth_gss.h
 *
 * Declarations for RPCSEC_GSS
 *
 * Copyright (c) 2000 Dug Song <dugsong@monkey.org>
 * Copyright (c) 2000 The Regents of the University of Michigan
 *
 * $Id: auth_gss.h,v 1.7 2000/08/24 19:32:07 dugsong Exp $
 */

#ifndef _LINUX_SUNRPC_AUTH_GSS_H
#define _LINUX_SUNRPC_AUTH_GSS_H

#ifdef __KERNEL__

/*
 * GSS-API definitions.
 */

typedef struct xdr_netobj	GSS_CTX_ID_T;
typedef struct xdr_netobj	GSS_CRED_ID_T;
typedef struct xdr_netobj	GSS_OID;
typedef struct xdr_netobj	GSS_BUFFER_T;

#define GSS_C_NO_BUFFER		((GSS_BUFFER_T) 0)
#define GSS_C_NO_CONTEXT	((GSS_CTX_ID_T) 0)
#define GSS_C_NULL_OID		((GSS_OID) 0)

#define GSS_C_QOP_DEFAULT	0

#define GSS_S_COMPLETE		0
#define GSS_S_CONTINUE_NEEDED	1

/*
 * RPCSEC_GSS definitions.
 */

#define RPC_GSS_VERSION		1

enum rpc_gss_proc {
	RPC_GSS_PROC_DATA = 0,
	RPC_GSS_PROC_INIT = 1,
	RPC_GSS_PROC_CONTINUE_INIT = 2,
	RPC_GSS_PROC_DESTROY = 3
};

#ifdef RPC_GSS_SVC_T_H
enum rpc_gss_svc {
	RPC_GSS_SVC_NONE = 1,
	RPC_GSS_SVC_INTEGRITY = 2,
	RPC_GSS_SVC_PRIVACY = 3
};
#endif 

struct rpc_gss_cred {
	u32		gc_v;		/* version */
	u32		gc_proc;	/* control procedure */
	u32		gc_seq;		/* sequence number */
	u32		gc_svc;		/* service */
	GSS_BUFFER_T	gc_ctx;		/* context handle */
};

struct rpc_gss_init_res {
	GSS_BUFFER_T	gr_ctx;		/* context handle */
	u32		gr_major;	/* major status */
	u32		gr_minor;	/* minor status */
	u32		gr_win;		/* sequence window */
	GSS_BUFFER_T	gr_token;	/* token */
};

/*
 * GSSD interface.
 */
void print_hexl(u32 *p, u_int length, u_int offset);

int gssd_client_test(struct xdr_netobj *name, u32 *uidp);
 
int gssd_name_to_uid(struct xdr_netobj *name, u32 *uidp);
int gssd_uid_to_name(u32 *uid, struct xdr_netobj *namep);

int gssd_name_to_gid(struct xdr_netobj *name, u32 *gidp);
int gssd_gid_to_name(u32 *gid, struct xdr_netobj *namep);

int gssd_init_sec_context(u32			uid,
			  u32		       *maj_statp,
			  u32		       *min_statp,
			  GSS_CTX_ID_T	       *ctx_id,
			  GSS_BUFFER_T	       *service_name,
			  GSS_OID	       *mech,
			  GSS_BUFFER_T	       *input_tokenp,
			  GSS_BUFFER_T	       *output_token,
			  u32		       *ret_flagsp);
     
int gssd_accept_sec_context(u32		       *maj_statp,
			    u32		       *min_statp,
			    GSS_CTX_ID_T       *ctx_id,
			    GSS_BUFFER_T       *input_token,
			    GSS_BUFFER_T       *client_namep,
			    GSS_BUFFER_T       *output_tokenp,
			    u32		       *ret_flagsp);

int
xdr_encode_rpc_gss_init_res(struct svc_rqst *req, u32 *p,
			    struct rpc_gss_init_res *resp);

#endif /* __KERNEL__ */
#endif /* _LINUX_SUNRPC_AUTH_GSS_H */

